#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <SDL.h>

#ifdef USE_OMP

#warning Using OpenMP in compilation: Make sure you remembered to pass -fopenmp to gcc.
#include <omp.h>

#endif

#include "model.h"

int mouse_x = 0, mouse_y = 0, mouse_button_down = 0;

int keys_held[64] = {0};

void remove_key( int keys_held[64], int v ) {
	int ii;
	for (ii=0; ii<64; ii++) {
		if (keys_held[ii] == v)
			keys_held[ii] = 0;
	}
}

void add_key( int keys_held[64], int v ) {
	int ii;
	remove_key( keys_held, v );
	for (ii=0; ii<64; ii++) {
		if (keys_held[ii] == 0) {
			keys_held[ii] = v;
			return;
		}
	}
}

int test_key( int keys_held[64], int v ) {
	int ii;
	for (ii=0; ii<64; ii++) {
		if (keys_held[ii] == v)
			return 1;
	}
	return 0;
}

#ifdef INTERLACED_RENDER
#ifndef INITIAL_INTERLACE_FACTOR
int INTERLACED_RENDER_factor = 1;
#else
int INTERLACED_RENDER_factor = INITIAL_INTERLACE_FACTOR;
#endif
#endif

#ifdef MOUSE_PERTURB
#ifndef INITIAL_MOUSE_EFFECT_STRENGTH
int mouse_effect_strength = 64;
#else
int mouse_effect_strength = INITIAL_MOUSE_EFFECT_STRENGTH;
#endif

#ifndef INITIAL_MOUSE_EFFECT_SIZE
int mouse_effect_size = 64;
#else
int mouse_effect_size = INITIAL_MOUSE_EFFECT_SIZE;
#endif

int controls_locked = 0;
int saved_mouse_x, saved_mouse_y, saved_mouse_button_down;
int saved_keys_held[64] = {0};
#endif

#ifdef OVERLAY_MOMENTUM
#ifndef INITIAL_MOMENTUM_UNIT_SIZE
int momentum_unit_size = 32;
#else
int momentum_unit_size = INITIAL_MOMENTUM_UNIT_SIZE;
#endif
#ifndef INITIAL_MOMENTUM_VISIBILITY
int momentum_visibility = 0;
#else
int momentum_visibility = INITIAL_MOMENTUM_VISIBILITY;
#endif
#endif

#ifdef PARTICLES_INVISIBLE
#ifndef INITIAL_PARTICLE_VISIBILITY
int particle_visibility = 1;
#else
int particle_visibility = INITIAL_PARTICLE_VISIBILITY;
#endif
#endif

int WIDTH, HEIGHT;

#define BPP 4
#define DEPTH 32

static void reset_layer(space_t* space, int layer);

static inline void setpixel(SDL_Surface *screen, int x, int y, Uint32 colour) {
	Uint32 *pixmem32;

//	if (x >= WIDTH) {
//		printf("%i\n", x - WIDTH);
//		return;
//	}
#ifdef LOW_PASS
	if (x < 0 || x >= WIDTH || y < 0 || y >= HEIGHT) return;
#endif

	y = y*screen->pitch/BPP;


	pixmem32 = (Uint32*) screen->pixels + y + x;
#ifndef LOW_PASS
	*pixmem32 = colour;
#else
	*pixmem32 += colour;
#endif
}

static inline void setpixel_bounds_check(SDL_Surface *screen, int x, int y, Uint32 colour) {
#ifndef LOW_PASS
	if (x < 0 || x >= WIDTH || y < 0 || y >= HEIGHT) return;
#endif
	setpixel(screen, x, y, colour);
}

void DrawScreen(SDL_Surface* screen, space_t* space) {

	if(SDL_MUSTLOCK(screen)) 
	{
		if(SDL_LockSurface(screen) < 0) return;
	}

#ifdef PARTICLES_INVISIBLE
#ifndef LOW_PASS
	if (!particle_visibility)
		SDL_FillRect(screen, NULL, SDL_MapRGB( screen->format, 0, 0, 0 ));
#endif
#endif

#ifdef LOW_PASS
#ifndef PSYCHEDELIC
	SDL_FillRect(screen, NULL, SDL_MapRGB( screen->format, 0, 0, 0 ));
#else
//	static int line_to_reset = 0;
//	line_to_reset += HEIGHT/7 + 1;
//	line_to_reset %= HEIGHT;
//	for (int ii=0; ii<HEIGHT/19; ii++) {
//		int line = (line_to_reset + ii) % HEIGHT;
//		memset(screen->pixels + BPP * WIDTH * line, 0, WIDTH*BPP);
//	}
	for (int ii=0; ii<WIDTH*HEIGHT*4; ii += 4) {
		((uint8_t*)screen->pixels)[ii] -= 8;
		((uint8_t*)screen->pixels)[ii+1] -= 3;
		((uint8_t*)screen->pixels)[ii+2] -= 8;
		((uint8_t*)screen->pixels)[ii+3] -= 7;
//		if (((uint8_t*)screen->pixels)[ii] < 50)
//			((uint8_t*)screen->pixels)[ii] = 0;
	}
	if ( ((uint8_t*)screen->pixels)[0] == 0 ) {
		for (int ii=0; ii<WIDTH*HEIGHT*4; ii += 4) {
			((uint8_t*)screen->pixels)[ii+2] = 0;
		}
		//SDL_FillRect(screen, NULL, SDL_MapRGB( screen->format, 0, 0, 0 ));
	}

#endif
#endif

#ifdef RESET_LAYERS
	static int layer_reset_timer = 0;
	static int layer_to_reset = 0;
	if ( layer_reset_timer++ == RESET_LAYERS ) {
		reset_layer(space, layer_to_reset++);
		layer_to_reset %= 6;
		layer_reset_timer = 0;
	}
#endif

	const cell_slice_t* restrict const read = space->state ? space->buffer1 : space->buffer0;

#ifdef INTERLACED_RENDER
	static int phase = 0;
	phase++;
	phase %= INTERLACED_RENDER_factor;
	phase %= HEIGHT;
#endif

	int true_yy_limit = space->height-1;
	if (true_yy_limit-1 > HEIGHT)
		true_yy_limit = HEIGHT+1;

#ifdef PARTICLES_INVISIBLE
	if (particle_visibility) {
#endif

#ifdef USE_OMP
	#pragma omp parallel shared(screen, space, true_yy_limit)
	{
		#pragma omp for
#endif

#ifdef INTERLACED_RENDER
	for (int yy=1+phase; yy<true_yy_limit; yy+=INTERLACED_RENDER_factor) {
#else
	for (int yy=1; yy<true_yy_limit; yy++) {
#endif
		for (int xx=1; xx<space->width-1; xx++) {
			cell_slice_t values = read[xx + yy * space->width];

			int loop_max = WORDLENGTH;
			if ((xx-1)*64 + WORDLENGTH > WIDTH) {
				loop_max = (WIDTH - (xx-1)*64);
			}

			for (int bit=0; bit<loop_max; bit++) {
				/*uint64_t sum = ( \
					(((uint64_t*)(p_values+7*0))[0] & 0x8000000000000000) | \
					(((uint64_t*)(p_values+7*1))[0] & 0x0080000000000000) | \
					(((uint64_t*)(p_values+7*2))[0] & 0x0000800000000000) | \
					(((uint64_t*)(p_values+7*3))[0] & 0x0000008000000000) | \
					(((uint32_t*)(p_values+7*4+4))[0] & 0x80000000) | \
					(((uint32_t*)(p_values+7*5+4))[0] & 0x00800000));
//*/
				//sum >>= 8;
				//sum += sum << 24;
				//sum += (sum << 16) + (sum << 8);
				//sum *= 0x010101010101;
				//sum >>= 56;

/*				int sum = \
					((uint64_t)(p_values[ 1]&0x80000000)) + \
					((uint64_t)(p_values[ 3]&0x80000000)) + \
					((uint64_t)(p_values[ 5]&0x80000000)) + \
					((uint64_t)(p_values[ 7]&0x80000000)) + \
					((uint64_t)(p_values[ 9]&0x80000000)) + \
					((uint64_t)(p_values[11]&0x80000000));*/
				int sum = (values.in[0]>>63) + (values.in[1]>>63) + (values.in[2]>>63) + (values.in[3]>>63) + (values.in[4]>>63) + (values.in[5]>>63);
				//int sum = (values.in[0]) + (values.in[1]) + (values.in[2]) + (values.in[3]) + (values.in[4]) + (values.in[5]);
				//int sum = 0;

				//int sum = ((values.in[0])&1) + ((values.in[1])&1) + ((values.in[2])&1) + ((values.in[3])&1) + ((values.in[4])&1) + ((values.in[5])&1);

				int xv = xx*WORDLENGTH + bit - 64;


#ifndef LOW_PASS
				uint32_t color = (sum*42);
#else
				uint32_t color = (sum*8);
#endif

#ifndef COLOR_FUNCTION
#define COLOR_FUNCTION color = color << 16 // By default, just red.
#endif

				COLOR_FUNCTION;

				setpixel(screen, xv, yy - 1, color );
#ifdef LOW_PASS
				setpixel(screen, xv-1, yy - 1, color );
				setpixel(screen, xv+1, yy - 1, color );
				setpixel(screen, xv, yy - 2,   color );
				setpixel(screen, xv, yy,       color );
#endif

				values.in[0] <<= 1;
				values.in[1] <<= 1;
				values.in[2] <<= 1;
				values.in[3] <<= 1;
				values.in[4] <<= 1;
				values.in[5] <<= 1;
			}
		}
	}
#ifdef USE_OMP
	}
#endif

#ifdef PARTICLES_INVISIBLE
	}
#endif 

#ifdef OVERLAY_MOMENTUM
	if (momentum_visibility) {

		double x_momenta[(space->height-2) / momentum_unit_size + 2][((space->width-2)*WORDLENGTH) / momentum_unit_size + 1];
		double y_momenta[(space->height-2) / momentum_unit_size + 2][((space->width-2)*WORDLENGTH) / momentum_unit_size + 1];
		memset(x_momenta, 0, sizeof(x_momenta));
		memset(y_momenta, 0, sizeof(y_momenta));

#define SIN_PI_OVER_4 0.70710678118654746

		for (int yy=1; yy<true_yy_limit; yy++) {
			for (int xx=1; xx<space->width-1; xx++) {
				cell_slice_t values = read[xx + yy * space->width];

				int loop_max = WORDLENGTH;
				if ((xx-1)*64 + WORDLENGTH > WIDTH) {
					loop_max = (WIDTH - (xx-1)*64);
				}

				for (int bit=0; bit<loop_max; bit++) {
					y_momenta[(yy-1) / momentum_unit_size][(xx*WORDLENGTH-bit-2) / momentum_unit_size] += (values.in[0]&1);
					x_momenta[(yy-1) / momentum_unit_size][(xx*WORDLENGTH-bit-2) / momentum_unit_size] -= (values.in[1]&1) * SIN_PI_OVER_4;
					y_momenta[(yy-1) / momentum_unit_size][(xx*WORDLENGTH-bit-2) / momentum_unit_size] += (values.in[1]&1) * SIN_PI_OVER_4;
					x_momenta[(yy-1) / momentum_unit_size][(xx*WORDLENGTH-bit-2) / momentum_unit_size] -= (values.in[2]&1);
					y_momenta[(yy-1) / momentum_unit_size][(xx*WORDLENGTH-bit-2) / momentum_unit_size] -= (values.in[3]&1);
					x_momenta[(yy-1) / momentum_unit_size][(xx*WORDLENGTH-bit-2) / momentum_unit_size] += (values.in[4]&1) * SIN_PI_OVER_4;
					y_momenta[(yy-1) / momentum_unit_size][(xx*WORDLENGTH-bit-2) / momentum_unit_size] -= (values.in[4]&1) * SIN_PI_OVER_4;
					x_momenta[(yy-1) / momentum_unit_size][(xx*WORDLENGTH-bit-2) / momentum_unit_size] += (values.in[5]&1);
					values.in[0] >>= 1;
					values.in[1] >>= 1;
					values.in[2] >>= 1;
					values.in[3] >>= 1;
					values.in[4] >>= 1;
					values.in[5] >>= 1;
				}
			}
		}

#ifdef DYNAMIC_ARROW_SIZING 
// Compute the laplacian, and use it to size the arrows.
		double laplacian[(space->height-2) / momentum_unit_size + 2][((space->width-2)*WORDLENGTH) / momentum_unit_size + 1];
		memset(laplacian, 0, sizeof(laplacian));

// Compute the mean laplacian, and drive arrows larger than it bigger, and arrows smaller than it smaller.
		double mean_laplacian = 0.0;
		int count = 0;

		for (int yy=1; yy<(space->height-2) / momentum_unit_size-1; yy++) {
			for (int xx=1; xx<((space->width-2)*WORDLENGTH) / momentum_unit_size - 1; xx++) {
				double local_laplacian = fabs(x_momenta[yy][xx+1] - x_momenta[yy][xx]) + fabs(x_momenta[yy][xx-1] - x_momenta[yy][xx]) + fabs(y_momenta[yy+1][xx] - y_momenta[yy][xx]) + fabs(y_momenta[yy-1][xx] - y_momenta[yy][xx]);
				laplacian[yy][xx] = local_laplacian;
				mean_laplacian += local_laplacian; // Numerical stability problems for large grids?
				count++;
			}
		}

		mean_laplacian *= 10;

		mean_laplacian /= count;

#endif

		for (int yy=0; yy<(space->height-2) / momentum_unit_size + 1; yy++) {
			for (int xx=0; xx<((space->width-2)*WORDLENGTH) / momentum_unit_size; xx++) {
#ifdef DYNAMIC_ARROW_SIZING
				// Skip cells marked as done.
				if (laplacian[yy][xx] == -1)
					continue;
#endif

				double x_momentum = x_momenta[yy][xx];
				double y_momentum = y_momenta[yy][xx];
				double x_offset = xx * momentum_unit_size + momentum_unit_size/2;
				double y_offset = yy * momentum_unit_size + momentum_unit_size/2;

				double norm, c;

#ifndef DYNAMIC_ARROW_SIZING
				norm = sqrt(x_momentum*x_momentum + y_momentum*y_momentum);
				c = 0.75 / (norm * momentum_unit_size);
#else

				// Note that we can only dynamically make arrows bigger, unless we do more computation to subdivide the momentum sums.
				if (laplacian[yy][xx] < mean_laplacian && ((xx&1)^1) && ((yy&1)^1) && (xx < ((space->width-2)*WORDLENGTH) / momentum_unit_size - 1) && ( yy < (space->height-2) / momentum_unit_size ) ) {
					if ((laplacian[yy][xx]+laplacian[yy][xx+1]+laplacian[yy+1][xx]+laplacian[yy+1][xx+1]) < mean_laplacian && ((xx&3) == 0) && ((yy&3) == 0) && (xx < ((space->width-2)*WORDLENGTH) / momentum_unit_size - 2) && ( yy < (space->height-2) / momentum_unit_size - 2 )) {
						x_momentum = 0; y_momentum = 0;
						for (int yyy=0; yyy<4; yyy++) {
							for (int xxx=0; xxx<4; xxx++) {
								laplacian[yy+yyy][xx+xxx] = -1;
								x_momentum += x_momenta[yy+yyy][xx+xxx];
								y_momentum += y_momenta[yy+yyy][xx+xxx];
							}
						}
						x_offset += momentum_unit_size/2 + momentum_unit_size;
						y_offset += momentum_unit_size/2 + momentum_unit_size;

						norm = sqrt(x_momentum*x_momentum + y_momentum*y_momentum);
						c = 0.75 / (norm * momentum_unit_size*4);
					} else {
						laplacian[yy][xx+1] = -1;
						laplacian[yy+1][xx]   = -1;
						laplacian[yy+1][xx+1] = -1;
						x_momentum += x_momenta[yy][xx+1] + x_momenta[yy+1][xx] + x_momenta[yy+1][xx+1];
						y_momentum += y_momenta[yy][xx+1] + y_momenta[yy+1][xx] + y_momenta[yy+1][xx+1];
						x_offset += momentum_unit_size/2;
						y_offset += momentum_unit_size/2;

						norm = sqrt(x_momentum*x_momentum + y_momentum*y_momentum);
						c = 0.75 / (norm * momentum_unit_size*2);
					}
				} else {
					norm = sqrt(x_momentum*x_momentum + y_momentum*y_momentum);
					c = 0.75 / (norm * momentum_unit_size);
				}
#endif

				int sx=0, sy=0;
				for (int ii=0; ii<norm; ii+=momentum_unit_size) {
					sx = x_offset + (x_momentum*ii) * c;
					sy = y_offset + (y_momentum*ii) * c;
					setpixel_bounds_check(screen, sx, sy, 0xff );
				}
				int arrow_size = norm*0.25;
				for (int ii=0; ii<arrow_size; ii+=momentum_unit_size) {
					for (int jj=0; jj<arrow_size-ii; jj+=momentum_unit_size) {
						setpixel_bounds_check(screen, sx - (y_momentum*ii - x_momentum*jj) * c, sy + (x_momentum*ii + y_momentum*jj) * c, 0xff00 );
						setpixel_bounds_check(screen, sx + (y_momentum*ii + x_momentum*jj) * c, sy - (x_momentum*ii - y_momentum*jj) * c, 0xff00 );
					}
				}
			}
		}

	}
#endif

#ifdef STREAM_VIDEO_TO_STDOUT
	write(1, screen->pixels, BPP * WIDTH * HEIGHT);
#endif

	if(SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);

	SDL_Flip(screen); 
}

static void reset_layer(space_t* space, int layer) {
	cell_slice_t* restrict const write = space->state ? space->buffer1 : space->buffer0;

	for (int yy=1; yy<space->height-1; yy++) {
		for (int xx=1; xx<space->width-1; xx++) {
			for (int ii=0; ii<6; ii++) {
#ifndef WIND_TUNNEL_BOUNDARY_CONDITIONS
#ifndef VACUUM
				if (layer == -1 || ii == layer) {
					// This anding and oring is the first few digits of the binary expansion of 1/6th, which is a density I decided on arbitrarily.
					write[xx + yy * space->width].in[ii] = (full_random() & full_random() & full_random()) | (full_random() & full_random() & full_random() & full_random() & full_random()) | (full_random() & full_random() & full_random() & full_random() & full_random() & full_random() & full_random());
				}
#endif
#else
				write[xx + yy * space->width].in[ii] = 0;
#endif

/*
				if (layer == -1 || ii == layer) {
					write[xx + yy * space->width].in[ii] = (full_random() & full_random() & full_random()) | (full_random() & full_random() & full_random() & full_random() & full_random()) | (full_random() & full_random() & full_random() & full_random() & full_random() & full_random() & full_random());
				}
				if (ii == 2)
					write[xx + yy * space->width].in[ii] = full_random() | full_random() | full_random();
*/
			}
		}
	}

#ifdef BOX_IN_THE_MIDDLE
	for (int yy=space->height/3; yy<(2*space->height)/3; yy++) {
		for (int xx=space->width/3; xx<(2*space->width)/3; xx++) {
			for (int ii=0; ii<6; ii++) {
				if (layer == -1 || ii == layer) {
					write[xx + yy * space->width].in[ii] = -1; //full_random();
				}
			}
		}
	}
#endif

}

static void reinit(space_t* space) {
	space->state = 0;

	memset(space->buffer0, 0, sizeof(cell_slice_t) * space->width * space->height);
	memset(space->buffer1, 0, sizeof(cell_slice_t) * space->width * space->height);

	reset_layer(space, -1);
}

int main(int argc, char **argv) {
	SDL_Surface *screen;
	SDL_Event event;
	const SDL_VideoInfo *info;
	struct timeval start_time;
	struct timeval stop_time;

	if (SDL_Init(SDL_INIT_VIDEO) < 0 ) return 1;

	info   = SDL_GetVideoInfo();
	if (info == NULL) {
		fprintf(stderr, "Unable to get video info: %s\n", SDL_GetError());
		return 1;
	}
	WIDTH  = info->current_w;
	HEIGHT = info->current_h;

#ifdef DROP_RESOLUTION
	WIDTH  /= DROP_RESOLUTION;
	HEIGHT /= DROP_RESOLUTION;
#endif

	fprintf(stderr, "Resolution: %ix%i\n", WIDTH, HEIGHT);

	int blocks_wide = (WIDTH+63)/64 + 2;
	int blocks_high = (HEIGHT+1)/64 + 1;
	fprintf(stderr, "Size: %ix%i blocks. (%ix%i active cells, %li in total.)\n", blocks_wide, blocks_high, blocks_wide*64-128, blocks_high*64-2, (long)(blocks_wide*64 - 128) * (long)(blocks_high*64 - 2));

	space_t space = allocate_space(blocks_wide, blocks_high);

	init_lfsr();

	reinit(&space);

#ifdef BENCHMARK_ONLY
	fprintf(stderr, "=== Benchmark Only Mode ===\nRecompile without -DBENCHMARK_ONLY=%i to recompile.\n", BENCHMARK_ONLY);
	gettimeofday(&start_time, NULL);

	for (int ii=0; ii<BENCHMARK_ONLY; ii++) {
		update(&space);
		// Swap the buffers;
		space.state ^= 1;
	}

	gettimeofday(&stop_time, NULL);

	long long elapsed_just_seconds  = stop_time.tv_sec - start_time.tv_sec;
	long long elapsed_just_useconds = stop_time.tv_usec - start_time.tv_usec;
	long long elapsed_useconds = elapsed_just_seconds*1000000 + elapsed_just_useconds;

	fprintf(stderr, "Run time:                        %.2f seconds\n", ((double)elapsed_useconds) / ((double)1000000));
	fprintf(stderr, "Updates completed:               %i\n", BENCHMARK_ONLY);
	fprintf(stderr, "Cell updates completed:          %lu\n", ((long)(blocks_wide*64 - 128) * (long)(blocks_high*64 - 2)) * (long)BENCHMARK_ONLY);
	fprintf(stderr, "Average updates per second:      %.2f \n", ((double)BENCHMARK_ONLY) * ((double)1000000) / ((double)elapsed_useconds));
	fprintf(stderr, "Average cell updates per second: %.2f million\n", ((double)BENCHMARK_ONLY) * ((double)((long)(blocks_wide*64 - 128) * (long)(blocks_high*64 - 2))) / ((double)elapsed_useconds));

	return 0;
}
#else

	if (!(screen = SDL_SetVideoMode(info->current_w, info->current_h, DEPTH, SDL_FULLSCREEN|SDL_HWSURFACE)))
	{
		SDL_Quit();
		return 1;
	}

#ifndef MOUSE_PERTURB
	SDL_ShowCursor(SDL_DISABLE);
#endif

	int continue_running   = 1;
	int total_update_count = 0;
	int total_frame_count  = 0;

	int rate = 0;
	int paused = 1;

#ifdef INITIAL_RATE
	rate = INITIAL_RATE;
#endif
#ifdef INITIAL_PAUSED
	paused = INITIAL_PAUSED;
#endif

	gettimeofday(&start_time, NULL);

	while(continue_running) {

#ifdef MOUSE_PERTURB
		if (!controls_locked) {
			saved_mouse_x = mouse_x;
			saved_mouse_y = mouse_y;
			saved_mouse_button_down = mouse_button_down;
			for (int ii=0; ii<64; ii++) {
				saved_keys_held[ii] = keys_held[ii];
			}
		}
#endif

		if (! paused) {

#ifdef MOUSE_PERTURB
			int modulus = mouse_effect_strength;
			int filter_length = 0;

			while (modulus > 1<<rate || filter_length < 1) {
				filter_length++;
				modulus >>= 1;
			}
#endif

			for (int ii=0; ii < 1<<rate; ii++) {
				update(&space);
				space.state ^= 1; // Swap the buffers;
				total_update_count++;

#ifdef MOUSE_PERTURB
				cell_slice_t* restrict const current = space.state ? space.buffer1 : space.buffer0;

				static int hits_count = 0;

// Waaaaah! If only macros were actually clean, or C were actually a powerful language!
#define test(key) int key_##key = test_key(saved_keys_held, SDLK_##key);
				test(LEFT)
				test(RIGHT)
				test(UP)
				test(DOWN)
				test(LSHIFT)
				test(RSHIFT)
#undef test

				if (saved_mouse_button_down && (hits_count++ % modulus == 0)) {
					for (int yy=-mouse_effect_size; yy<mouse_effect_size; yy++) {
						for (int xx=-mouse_effect_size; xx<mouse_effect_size; xx++) {
							int x = xx + saved_mouse_x + WORDLENGTH;
							int y = yy + saved_mouse_y + 1;
							if (x < 0 || x >= WIDTH || y < 0 || y >= HEIGHT) continue;
							for (int ii=0; ii<6; ii++) {
								bitslice rb = 1;
								int jj = filter_length;
								while (jj--)
									rb &= update_random();
								// Determine the type of command being given.
								if (key_LEFT) {
									if (ii == 2)
										current[x/WORDLENGTH + y * space.width].in[ii] |= (bitslice)rb << (WORDLENGTH - 1 - x%WORDLENGTH);
									else
										current[x/WORDLENGTH + y * space.width].in[ii] &= ~ ( (bitslice)rb << (WORDLENGTH - 1 - x%WORDLENGTH) );
								} else if (key_RIGHT) {
									if (ii == 5)
										current[x/WORDLENGTH + y * space.width].in[ii] |= (bitslice)rb << (WORDLENGTH - 1 - x%WORDLENGTH);
									else
										current[x/WORDLENGTH + y * space.width].in[ii] &= ~ ( (bitslice)rb << (WORDLENGTH - 1 - x%WORDLENGTH) );
								} else if (key_UP) {
									if (ii == 3)
										current[x/WORDLENGTH + y * space.width].in[ii] |= (bitslice)rb << (WORDLENGTH - 1 - x%WORDLENGTH);
									else
										current[x/WORDLENGTH + y * space.width].in[ii] &= ~ ( (bitslice)rb << (WORDLENGTH - 1 - x%WORDLENGTH) );
								} else if (key_DOWN) {
									if (ii == 0)
										current[x/WORDLENGTH + y * space.width].in[ii] |= (bitslice)rb << (WORDLENGTH - 1 - x%WORDLENGTH);
									else
										current[x/WORDLENGTH + y * space.width].in[ii] &= ~ ( (bitslice)rb << (WORDLENGTH - 1 - x%WORDLENGTH) );
								} else if (key_LSHIFT || key_RSHIFT) {
									current[x/WORDLENGTH + y * space.width].in[ii] |= (bitslice)rb << (WORDLENGTH - 1 - x%WORDLENGTH);
								}
							}
						}
					}
				}
#endif
			}
		}

		DrawScreen(screen, &space);
		total_frame_count++;

		while(SDL_PollEvent(&event))
		{
			switch (event.type)
			{
				case SDL_QUIT:
					continue_running = 0;
					break;
				case SDL_MOUSEBUTTONDOWN:
					if( event.button.button == SDL_BUTTON_LEFT ) {
						mouse_button_down = 1;
					}
					break;
				case SDL_MOUSEBUTTONUP:
					if ( event.button.button == SDL_BUTTON_LEFT ) {
						mouse_button_down = 0;
					}
					break;
				case SDL_KEYUP:
					remove_key(keys_held, event.key.keysym.sym);
					break;
				case SDL_KEYDOWN:
					add_key(keys_held, event.key.keysym.sym);
					if (event.key.keysym.sym == '=') {
						rate++;
					}
					if (event.key.keysym.sym == '-') {
						rate--;
					}
#ifdef INTERLACED_RENDER
					if (event.key.keysym.sym == '1')
						INTERLACED_RENDER_factor /= 2;
					if (event.key.keysym.sym == '2')
						INTERLACED_RENDER_factor *= 2;
					if (INTERLACED_RENDER_factor < 1)
						INTERLACED_RENDER_factor = 1;
					if ((test_key(keys_held, SDLK_LCTRL) || test_key(keys_held, SDLK_RCTRL)) && event.key.keysym.sym == '1') {
						rate = 0;
						INTERLACED_RENDER_factor = 1;
					}
#endif
#ifdef MOUSE_PERTURB
					if (event.key.keysym.sym == SDLK_DELETE)
						controls_locked ^= 1;
					if (event.key.keysym.sym == '6') {
						if (test_key(keys_held, SDLK_LCTRL) || test_key(keys_held, SDLK_RCTRL))
							saved_mouse_button_down ^= 1;
						else
							mouse_effect_strength *= 2;
					}
					if (event.key.keysym.sym == '7')
						mouse_effect_strength /= 2;
					if (mouse_effect_strength < 4)
						mouse_effect_strength = 4;
					if (event.key.keysym.sym == '8') {
						if (test_key(keys_held, SDLK_LCTRL) || test_key(keys_held, SDLK_RCTRL))
							mouse_effect_size = 1;
						else
							mouse_effect_size /= 2;
					}
					if (event.key.keysym.sym == '9')
						mouse_effect_size *= 2;
					if (mouse_effect_strength < 1)
						mouse_effect_strength = 1;
#endif
#ifdef OVERLAY_MOMENTUM
					if (event.key.keysym.sym == '3')
						momentum_unit_size /= 2;
					if (event.key.keysym.sym == '4') {
						if (test_key(keys_held, SDLK_LCTRL) || test_key(keys_held, SDLK_RCTRL))
							momentum_visibility ^= 1;
						else
							momentum_unit_size *= 2;
					}
					if ((test_key(keys_held, SDLK_LCTRL) || test_key(keys_held, SDLK_RCTRL)) && event.key.keysym.sym == '3')
						momentum_unit_size = 0;
					if (momentum_unit_size < 4)
						momentum_unit_size = 4;
#endif
#ifdef PARTICLES_INVISIBLE
					if (event.key.keysym.sym == '5')
						particle_visibility ^= 1;
#endif
					if (event.key.keysym.sym == SDLK_SPACE) {
						paused ^= 1;
					}
					if (event.key.keysym.sym == SDLK_RETURN) {
						time_reverse(&space);
					}
					if (event.key.keysym.sym == SDLK_BACKSPACE) {
						reinit(&space);
#ifdef AUTO_EQUILIBRIATE
						for (int ii=0; ii<AUTO_EQUILIBRIATE; ii++) {
							update(&space);
							space.state ^= 1; // Swap the buffers;
							total_update_count++;
						}
#endif
					}
					if (event.key.keysym.sym == SDLK_LSHIFT)
						scattering ^= 1;
					if (rate < 0) rate = 0;
					if (event.key.keysym.sym == SDLK_ESCAPE) {
						continue_running = 0;
					}
					break;
				case SDL_MOUSEMOTION:
					mouse_x = event.button.x;
					mouse_y = event.button.y;
					break;
			}
		}
	}

	gettimeofday(&stop_time, NULL);

	long long elapsed_just_seconds  = stop_time.tv_sec - start_time.tv_sec;
	long long elapsed_just_useconds = stop_time.tv_usec - start_time.tv_usec;
	long long elapsed_useconds = elapsed_just_seconds*1000000 + elapsed_just_useconds;

	fprintf(stderr, "Run time:                        %.2f seconds\n", ((double)elapsed_useconds) / ((double)1000000));
	fprintf(stderr, "Updates completed:               %i\n", total_update_count);
	fprintf(stderr, "Frames completed:                %i\n", total_frame_count);
	fprintf(stderr, "Cell updates completed:          %lu\n", ((long)(blocks_wide*64 - 128) * (long)(blocks_high*64 - 2)) * (long)total_update_count);
#ifdef COUNT_RANDOM_BITS_FOR_FUN
	fprintf(stderr, "LFSR steps computed:             %lu\n", random_bits_consumed);
#endif
	fprintf(stderr, "Average updates per second:      %.2f \n", ((double)total_update_count) * ((double)1000000) / ((double)elapsed_useconds));
	fprintf(stderr, "Average frames per second:       %.2f \n", ((double)total_frame_count) * ((double)1000000) / ((double)elapsed_useconds));
	fprintf(stderr, "Average cell updates per second: %.2f million\n", ((double)total_update_count) * ((double)((long)(blocks_wide*64 - 128) * (long)(blocks_high*64 - 2))) / ((double)elapsed_useconds));
#ifdef COUNT_RANDOM_BITS_FOR_FUN
	fprintf(stderr, "Average LFSR steps per second:   %.2f million\n", ((double)random_bits_consumed) / ((double)elapsed_useconds) );
#endif

	SDL_Quit();

	return 0;
}

#endif // Benchmark only.

