// Lattice gas.

#include "model.h"

int scattering = 0;

#ifdef USE_OMP

#warning Using OpenMP in compilation: Make sure you remembered to pass -fopenmp to gcc.
#include <omp.h>

#endif

space_t allocate_space(int width, int height) {
	space_t result;

	height *= WORDLENGTH;

	result.width   = width;
	result.height  = height;
	result.state = 0;
	result.buffer0 = (cell_slice_t*) malloc( sizeof(cell_slice_t) * width * height );
	result.buffer1 = (cell_slice_t*) malloc( sizeof(cell_slice_t) * width * height );

	memset(result.buffer0, 0, sizeof(cell_slice_t) * width * height);
	memset(result.buffer1, 0, sizeof(cell_slice_t) * width * height);

	return result;
}

uint64_t lfsr;

#define LFSR_STEP(x) (((x) >> 1) ^ (uint64_t)((0 - ((x) & (uint64_t)1)) & 0x800000000000000d))

#ifdef COUNT_RANDOM_BITS_FOR_FUN
long random_bits_consumed = 0;
#endif

uint64_t update_random(void) {

#ifdef COUNT_RANDOM_BITS_FOR_FUN
	random_bits_consumed++;
#endif

	lfsr = LFSR_STEP(lfsr);
	return lfsr;
}

uint64_t full_random(void) {
	for (int ii=0; ii<64; ii++)
		update_random();
	return lfsr;
}

void init_lfsr(void) {
	lfsr = time(NULL);
	for (int ii=0; ii<1000; ii++) {
		update_random();
	}
}

void update(const space_t* restrict const s) {
	cell_slice_t* restrict const read  = s->state ? s->buffer1 : s->buffer0;
	cell_slice_t* restrict const write = s->state ? s->buffer0 : s->buffer1;

#ifdef USE_OMP
	uint64_t random_seed = full_random();
	#pragma omp parallel
	{
		uint64_t temp_lfsr = random_seed ^ omp_get_thread_num();
#ifdef COUNT_RANDOM_BITS_FOR_FUN
		int bits_used_by_this_thread = 64;
#endif
		for (int ii=0; ii<64; ii++)
			temp_lfsr = LFSR_STEP(temp_lfsr);
		#pragma omp for
#endif
	for (int yy=1; yy<s->height-1; yy++) {
		for (int xx=1; xx<s->width-1; xx++) {
#ifndef NO_RANDOMNESS
#ifndef USE_OMP
			bitslice random_bits = update_random();
#else
			bitslice random_bits = temp_lfsr;
			temp_lfsr = LFSR_STEP(temp_lfsr);

#ifdef COUNT_RANDOM_BITS_FOR_FUN
			bits_used_by_this_thread++;
#endif

#endif
#else
			bitslice random_bits = 0;
#endif

			cell_slice_t inputs = read[xx + yy * s->width];

			// By default, transport.
			cell_slice_t outputs = inputs;

#ifndef HEAD_ON_PASSES_THROUGH
			// Model head on collisions.
			const bitslice collision0 =  inputs.in[0] & ~inputs.in[1] & ~inputs.in[2] &  inputs.in[3] & ~inputs.in[4] & ~inputs.in[5];
			const bitslice collision1 = ~inputs.in[0] &  inputs.in[1] & ~inputs.in[2] & ~inputs.in[3] &  inputs.in[4] & ~inputs.in[5];
			const bitslice collision2 = ~inputs.in[0] & ~inputs.in[1] &  inputs.in[2] & ~inputs.in[3] & ~inputs.in[4] &  inputs.in[5];

			outputs.in[0] &= ~collision0; outputs.in[3] &= ~collision0;
			outputs.in[1] |= collision0 & random_bits;  outputs.in[4] |= collision0 & random_bits;
			outputs.in[2] |= collision0 & ~random_bits; outputs.in[5] |= collision0 & ~random_bits;

			outputs.in[1] &= ~collision1; outputs.in[4] &= ~collision1;
			outputs.in[2] |= collision1 & random_bits;  outputs.in[5] |= collision1 & random_bits;
			outputs.in[3] |= collision1 & ~random_bits; outputs.in[0] |= collision1 & ~random_bits;

			outputs.in[2] &= ~collision2; outputs.in[5] &= ~collision2;
			outputs.in[3] |= collision2 & random_bits;  outputs.in[0] |= collision2 & random_bits;
			outputs.in[4] |= collision2 & ~random_bits; outputs.in[1] |= collision2 & ~random_bits;
#endif

			// Model three way collisions.
			const bitslice collision3 =  inputs.in[0] & ~inputs.in[1] &  inputs.in[2] & ~inputs.in[3] &  inputs.in[4] & ~inputs.in[5];
			const bitslice collision4 = ~inputs.in[0] &  inputs.in[1] & ~inputs.in[2] &  inputs.in[3] & ~inputs.in[4] &  inputs.in[5];

#ifndef THREE_WAY_RANDOMLY_SCATTERS
			outputs.in[0] &= ~collision3; outputs.in[2] &= ~collision3; outputs.in[4] &= ~collision3;
			outputs.in[1] |= collision3;  outputs.in[3] |= collision3;  outputs.in[5] |= collision3;

			outputs.in[1] &= ~collision4; outputs.in[3] &= ~collision4; outputs.in[5] &= ~collision4;
			outputs.in[0] |= collision4;  outputs.in[2] |= collision4;  outputs.in[4] |= collision4;
#else
			outputs.in[0] &= ~collision3 | random_bits; outputs.in[2] &= ~collision3 | random_bits; outputs.in[4] &= ~collision3 | random_bits;
			outputs.in[1] |= collision3 & ~random_bits;  outputs.in[3] |= collision3 & ~random_bits;  outputs.in[5] |= collision3 & ~random_bits;

			outputs.in[1] &= ~collision4 | random_bits; outputs.in[3] &= ~collision4 | random_bits; outputs.in[5] &= ~collision4 | random_bits;
			outputs.in[0] |= collision4 & ~random_bits;  outputs.in[2] |= collision4 & ~random_bits;  outputs.in[4] |= collision4 & ~random_bits;
#endif

			// Push outputs
			write[xx   + (yy+1) * s->width].in[0]  = outputs.in[0];
			write[xx   + (yy+1) * s->width].in[1]  = outputs.in[1] << 1;
			write[xx-1 + (yy+1) * s->width].in[1] |= outputs.in[1] >> 63;
			write[xx   +  yy    * s->width].in[2] |= outputs.in[2] << 1;
			write[xx-1 +  yy    * s->width].in[2] |= outputs.in[2] >> 63;
			write[xx   + (yy-1) * s->width].in[3]  = outputs.in[3];
			write[xx   + (yy-1) * s->width].in[4] |= outputs.in[4] >> 1;
			write[xx+1 + (yy-1) * s->width].in[4] |= outputs.in[4] << 63;
			write[xx   +  yy    * s->width].in[5] |= outputs.in[5] >> 1;
			write[xx+1 +  yy    * s->width].in[5]  = outputs.in[5] << 63;
		}
	}
#ifdef USE_OMP
	}
#endif

#ifndef WIND_TUNNEL_BOUNDARY_CONDITIONS
	// Enforce periodic boundary conditions:

	// Top and bottom rows
	for (int xx=0; xx<s->width; xx++) {
		for (int ii=0; ii<6; ii++) {
			write[xx + (s->height-2) * s->width].in[ii] ^= write[xx].in[ii];
		}
		for (int ii=0; ii<6; ii++) {
			write[xx + s->width].in[ii] ^= write[xx + (s->height-1) * s->width].in[ii];
		}
	}

	// Left and right sides.
	for (int yy=0; yy<s->height; yy++) {
		for (int ii=0; ii<6; ii++) {
			write[s->width-2 + yy * s->width].in[ii] ^= write[yy * s->width].in[ii];
		}
		for (int ii=0; ii<6; ii++) {
			write[1 + yy * s->width].in[ii] ^= write[s->width-1 + yy * s->width].in[ii];
		}
	}

#else // Wind tunnel boundary conditions!

	// Right side, random particles.
	for (int yy=0; yy<s->height; yy++) {
		for (int ii=0; ii<6; ii++) {
			write[s->width-2 + yy * s->width].in[ii] ^= update_random() & update_random() & 1;
			if (ii == 2)
				write[s->width-2 + yy * s->width].in[ii] ^= update_random() & 1;
		}
	}

	// Top and bottom rows, source random particles
	for (int xx=0; xx<s->width; xx++) {
		for (int ii=0; ii<6; ii++) {
			write[xx + (s->height-2) * s->width].in[ii] = update_random() & update_random();
		}
		for (int ii=0; ii<6; ii++) {
			write[xx + s->width].in[ii] = update_random() & update_random();
		}
	}

#endif

	memset(read, 0, sizeof(*read) * s->width * s->height);

}

void time_reverse(const space_t* restrict s) {
	cell_slice_t* restrict const read  = s->state ? s->buffer1 : s->buffer0;

	for (int yy=1; yy<s->height-1; yy++) {
		for (int xx=1; xx<s->width-1; xx++) {
			cell_slice_t inputs = read[xx + yy * s->width];
			read[xx + yy * s->width].in[0] = inputs.in[3];
			read[xx + yy * s->width].in[1] = inputs.in[4];
			read[xx + yy * s->width].in[2] = inputs.in[5];
			read[xx + yy * s->width].in[3] = inputs.in[0];
			read[xx + yy * s->width].in[4] = inputs.in[1];
			read[xx + yy * s->width].in[5] = inputs.in[2];
		}
	}
}

